/*
 * The Spar Library - modular math parser
 * Copyright (C) 2000,2001 Davide Angelocola <davide178@inwind.it>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 */
// TODO: add more detailed domain error messages
// TODO: checks to domain

#include <math.h>
#include <spar/sl_math.h>
#include <spar/sl_error.h>
#include <spar/sl_math_library.h>

SL_MATH_FUNCTION (sl_deg)
{
  return SL_TO_DEG (x);
}

SL_MATH_FUNCTION (sl_rad)
{
  return SL_TO_RAD (x);
}

SL_MATH_FUNCTION_2 (sl_sqrn)
{
  if (SL_IS_ZERO (x))
    return ZERO;

  if (((int) y % 2) == 0)
    {
      if (SL_IS_NEGATIVE (x))
	sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_NEGARG, "");
      else
	return (exp (sl_inv (y) * sl_log (x)));
    }

  if (SL_IS_POSITIVE (x))
    return (exp (sl_inv (y) * sl_log (x)));
  else
    return -(exp (sl_inv (y) * sl_log (-x)));
}

SL_MATH_FUNCTION (sl_log)
{
  if (SL_IS_NEGATIVE (x))
    sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_NEGARG, "");

  if (SL_IS_ZERO (x))
    sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_NULLARG, "");

  return (log (x));
}


SL_MATH_FUNCTION (sl_log10)
{
  if (SL_IS_NEGATIVE (x))
    sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_NEGARG, "");

  if (SL_IS_ZERO (x))
    sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_NULLARG, "");

  return (log10 (x));
}

SL_MATH_FUNCTION (sl_arcsin)
{
  if ((x > 1) || (x < -1))
    sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_DOMAIN, "");

  return asin (x);
}


SL_MATH_FUNCTION (sl_arccos)
{
  if ((x > 1) || (x < -1))
    sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_DOMAIN, "");

  return acos (x);
}

SL_MATH_FUNCTION (sl_arctan)
{
  return atan (x);
}

SL_MATH_FUNCTION (sl_sin)
{
  return sin (x);
}

SL_MATH_FUNCTION (sl_cos)
{
  return cos (x);
}

SL_MATH_FUNCTION (sl_tan)
{
  /* todo: checks the domain */
  return tan (x);
}

SL_MATH_FUNCTION (sl_sinh)
{
  return sinh (x);
}

SL_MATH_FUNCTION (sl_cosh)
{
  return cosh (x);
}

SL_MATH_FUNCTION (sl_tanh)
{
  return tanh (x);
}

SL_MATH_FUNCTION (sl_exp)
{
  if (SL_IS_ZERO (x))
    return 1.0;
  else
    return (exp (x));
}

SL_MATH_FUNCTION_2 (sl_expn)
{
  return exp (x * sl_log (y));
}

SL_MATH_FUNCTION (sl_floor)
{
  if (SL_IS_ZERO (x))
    return ZERO;
  else
    return floor (x);
}

SL_MATH_FUNCTION (sl_ceil)
{
  if (SL_IS_ZERO (x))
    return ZERO;
  else
    return ceil (x);
}

SL_MATH_FUNCTION (sl_abs)
{
  if (SL_IS_ZERO (x))
    return ZERO;
  else
    return fabs (x);
}

SL_MATH_FUNCTION (sl_fat)
{
  register double f = 1.0;
  register int i;

  if (x <= 1.0)
    return 1.0;
  else
    {
      for (i = 1; i <= x; i++)
	f = f * (double) i;

      return f;
    }
}

SL_MATH_FUNCTION (sl_sum)
{
  if (x < 0)
    sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_NEGARG, "");

  return ((x + 1) * (x / 2));
}

SL_MATH_FUNCTION_2 (sl_max)
{
  return SL_MAX (x, y);
}

SL_MATH_FUNCTION_2 (sl_min)
{
  return SL_MIN (x, y);
}

SL_MATH_FUNCTION (sl_inv)
{
  if (SL_IS_ZERO (x))
    sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_DIVZERO, "");

  return (1 / x);
}


SL_MATH_FUNCTION (sl_log2)
{
  if (SL_IS_NEGATIVE (x))
    sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_NEGARG, "");

  if (SL_IS_ZERO (x))
    sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_NULLARG, "");

  return (sl_log (x) / M_LN2);
}

SL_MATH_FUNCTION (sl_sgn)
{
  if (SL_IS_NEGATIVE (x))
    return -1.0;
  else if (SL_IS_POSITIVE (x))
    return 1.0;
  else
    return ZERO;
}

SL_MATH_FUNCTION_2 (sl_pow)
{
  return pow (x, y);
}

SL_MATH_FUNCTION_2 (sl_mod)
{
  if (SL_IS_ZERO (x))
    sl_error_throw (SL_ERROR_CLASS_PARSER, SL_ERROR_DIVZERO, "");

  return fmod (x, y);
}

SL_MATH_FUNCTION_2 (sl_mcd)
{
  double rem;
  double a = x;
  double b = y;
  while (SL_IS_NOT_ZERO (y))
    {
      rem = fmod (a, b);
      a = b;
      b = rem;
    }

  return a;
}
