/*
 * Spar - a text frontend for the Spar library
 * Copyright (C) 2000,2001 Davide Angelocola <davide178@inwind.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <spar/sl.h>

/*
 *  If this system doesn't have the getopt library, 
 *  try to compile the getopt.c file and link it
 */ 
#ifndef HAVE_GETOPT_H
#warning System include file "getopt.h" not found.
#warning 
#include "getopt.h"
#else
#include <getopt.h>
#endif
 
#include "terminal.h"
#include "prefs.h"

static void
welcome (void)
{
  sl_writeln ("This is free software with ABSOLUTELY NO WARRANTY.");
  sl_writeln ("Type 'help' for a list of commands");
  sl_writeln ("");

  return;
}

void
clean_exit (int exit_code)
{
#ifdef HAVE_READLINE_READLINE_H
#ifdef HAVE_READLINE_HISTORY_H
  extern void free_history (void);
  free_history ();
#endif /* HAVE_READLINE_HISTORY_H */
#endif /* HAVE_READLINE_READLINE_H */

  sl_de_init ();
  restore_term ();
  sl_writeln ("");
  exit (exit_code);
}


extern int init_commands (void);
/*
 * Spar "shell" extern reference
 */
extern void shell (void);

/* 
 * Getopt Library Interface
 */
extern int optind;
extern char *optarg;

static struct option long_options[] = {
  {"quiet", 0, 0, 'q'},
  {"help", 0, 0, 'h'},
  {"decimals", 0, 0, 'd'},
  {"integers", 0, 0, 'i'},
  {"config", 0, 0, 'c'},
  {"version", 0, 0, 'v'},
  {0, 0, 0, 0}
};

static void
usage (void)
{
  sl_write ("%s ", PACKAGE); sl_version ();
  sl_writeln ("Copyright (C) 2000,2001 Davide Angelocola");
  sl_writeln ("This is free software with ABSOLUTELY NO WARRANTY.");
  sl_writeln ("");
  sl_writeln ("spar [options] expression");
  sl_writeln ("");
  sl_writeln ("Options:");
  sl_writeln ("  --help       show this message");
  sl_writeln ("  --version    show program version");
  sl_writeln ("  --config     specify a user modules config file");
  sl_writeln ("  --quiet      don't show welcome message");
  sl_writeln ("");
  sl_writeln ("Precision:");
  sl_writeln ("  --decimals   the number of decimals digits");
  sl_writeln ("  --integers   the number of integers digits");
  sl_writeln ("");
  sl_writeln
    ("Additional information about Spar is avaiable via the WWW at:");
  sl_writeln ("http://spar.sourceforge.net (main site)");
  sl_writeln ("http://web.tiscalinet.it/Davide18/index.html");
  sl_writeln ("");
  sl_writeln ("Please report bugs to <davide178@inwind.it>");

  exit (1);
}

static void
version (void)
{
  sl_info ();
  exit (1);
}

/*
 * preference structure (see prefs.h)
 * this the only istance
 * loaded with the defaults
 */
struct preferences prefs = {
  max_history_entries:256,
  verbose_level:1,		// not used yet
  prec_decimal:0,
  prec_integer:0,
  prompt:"# ",
  modules_config_file:SL_MODULE_CONFIG_FILE
};


int
main (int argc, char **argv)
{
  char c;
  int option_index = 0, quiet_flag = 0;

  while (1)
    {
      c = getopt_long (argc, argv, "i:d:c:qhv", long_options, &option_index);

      if (c == -1)
	break;

      switch (c)
	{
	case 'h':
	  usage ();
	  break;

	case 'v':
	  version ();
	  break;

	case 'c':
	  strcpy (prefs.modules_config_file, optarg);
	  break;

	case 'q':
	  quiet_flag = 1;
	  break;

	case 'd':
	  {
	    int d = atoi (optarg);
	    if (d > 0)
	      prefs.prec_decimal = d;
	    break;
	  }
	case 'i':
	  {
	    int i = atoi (optarg);
	    if (i > 0)
	      prefs.prec_decimal = i;
	    break;
	  }
	default:
	  exit (1);
	}
    }

  /*
   * Evaluate command line 
   * i.e. spar cos(90)
   */
  if (optind < argc)
    {
      char expression[1024] = "";

      while (optind < argc)
	strcat (expression, argv[optind++]);

      sl_parser_evaluate (expression);

      exit (0);
    }

  /*
   * Inits some terminal parameter such as:  
   *   inibiths the use of CTRL-D key
   *   installs signals handlers
   *   save the old configuration to a termios struct
   *   other minor setting
   */
  init_term ();

  /*
   * Inits the Spar Library 
   */
  sl_init ();

  /* 
   * Load all modules in the configuration file
   * This file must be in your home directory
   */
  sl_module_loader_load_config (get_preference (modules_config_file));

  /* 
   * Set the precision in which ways the number are printed 
   */
  sl_writer_set_precision (get_preference (prec_decimal),
			   get_preference (prec_integer));

  init_commands ();

  /* 
   * Prints a welcome message 
   */
  if (!quiet_flag)
    welcome ();

  /*
   * Spar "shell"
   */
  shell ();

  /* not reached */
  return 0;
}
