/*
 * Spar - a text frontend for the Spar library
 * Copyright (C) 2000,2001 Davide Angelocola <davide178@inwind.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

// ToDo: ! call last command

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <spar/sl_parser.h>
#include <spar/sl_memory.h>
#include <spar/sl_writer.h>

#include "prefs.h"
#include "commands_table.h"

extern const int COMMANDS_COUNTER;

#ifdef HAVE_READLINE_HISTORY_H
#include <readline/history.h>

int
show_history (int a, int b)
{
  /* this is the reference to history table */
  HIST_ENTRY **hlist;
  register int i;

  /* gets history table */
  hlist = history_list ();

  /* check if the history is empty */
  if (hlist)
    {
      /* print history table */
      sl_writeln ("\n");
      for (i = 0; hlist[i]; i++)
	sl_writeln ("%d: %s", i + history_base, hlist[i]->line);
    }
  else
    {
      /* history is empty */
      sl_writeln ("no history entries.");
    }

  sl_writeln ("[%d,%d]", history_length, max_input_history);
  sl_writeln ("");
  sl_write ("%s", (get_preference (prompt)));

  return 1;
}

void
add_to_history (char *line)
{
  if (line && *line)
    add_history (line);
}

void
free_history (void)
{
  HIST_ENTRY **list = history_list ();
  register int i;

  if (list)
    {
      for (i = 0; list[i]; i++)
	sl_free (list[i]->line, strlen (list[i]->line));

      sl_free (list, sizeof (**list));
    }
}
#endif /* HAVE_READLINE_HISTORY_H */


#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>

char *
dupstr (char *s)
{
  char *r = (char *) sl_malloc (strlen (s) + 1);
  strcpy (r, s);
  return r;
}

/*
 * Generator function for command completion.  STATE lets us know whether
 * to start from scratch; without any state (i.e. STATE == 0), then we
 * start at the top of the list.
 */
char *
command_generator (char *text, int state)
{
  static int i, len;

  /*
   * If this is a new word to complete, initialize now.  This includes
   * saving the length of TEXT for efficiency, and initializing the index
   * variable to 0.
   */
  if (!state)
    {
      i = 0;
      len = strlen (text);
    }

  while (i < COMMANDS_COUNTER)
    {
      i++;
      if ((strncmp (Commands[i - 1].cmd, text, len)) == 0)
	return (dupstr (Commands[i - 1].cmd));
    }

  /*
   * Return the next name which partially matches from the command list.
   */
  return ((char *) NULL);
}

/*
 * Attempt to complete on the contents of TEXT.  START and END bound the
 * region of rl_line_buffer that contains the word to complete.  TEXT is
 * the word to complete.  We can use the entire contents of rl_line_buffer
 * in case we want to do some simple parsing.  Return the array of matches\,
 * or NULL if there aren't any.
 */
char **
spar_completion (char *text, int start, int end)
{
  char **matches;

  matches = (char **) NULL;

  /* If this word is at the start of the line, then it is a command
     to complete.  Otherwise it is the name of a file in the current
     directory. */
  if (start == 0)
    matches = completion_matches (text, (CPFunction *) command_generator);

  return (matches);
}

/* Read a string, and return a pointer to it.  Returns NULL on EOF. */
char *
rl_gets (void)
{
  static char *line_read = (char *) NULL;
  /* If the buffer has already been allocated, return the memory
     to the free pool. */
  if (line_read)
    {
      sl_free (line_read, strlen (line_read));
      line_read = (char *) NULL;
    }

  line_read = readline (get_preference (prompt));

  if (strlen (line_read) > 1024)
    {
      /* line entered is too large...return a blank string */
      sl_writeln ("line is too large. ");
      return "";
    }
  else
    {
#ifdef HAVE_READLINE_HISTORY_H
      /* If the line has any text in it, save it on the history. */
      extern void add_to_history (char *);
      add_to_history (line_read);
#endif /* HAVE_READLINE_HISTORY_H */
      /* return this to parse_cmd_line */
      return line_read;
    }
}

#endif /* HAVE_READLINE_READLINE_H */

void
initialize_readline_history (void)
{
#ifdef HAVE_READLINE_READLINE_H
  /* Allow conditional parsing of the ~/.inputrc file. */
  rl_readline_name = "Spar";
  rl_basic_word_break_characters = "+-*/; ^|&<>=?:(),";
  rl_attempted_completion_function = (CPPFunction *) spar_completion;
#endif /* HAVE_READLINE_READLINE_H */
#ifdef HAVE_READLINE_HISTORY_H

  /* setting some history parameters */
  using_history ();
  stifle_history (get_preference (max_history_entries));

#ifdef HAVE_READLINE_READLINE_H
  rl_bind_key ('\\', show_history);
#endif /* HAVE_READLINE_READLINE_H */
#endif /* HAVE_READLINE_HISTORY_H */
}

#define INPUT_LINE_BUFFER_LENGHT	256
#define MAX_COMMAND_ARGS		64

static int parse_cmdline (char *, char **);

/*
 *	Run commands and parse expressions
 *	think this funciotns as a small shell
 *
 */
void
shell (void)
{
  register int command_number, argc;
  char *argv[MAX_COMMAND_ARGS];
#ifdef HAVE_READLINE_READLINE_H
  char *line = 0;
#else
  char line[INPUT_LINE_BUFFER_LENGHT];
#endif /* HAVE_READLINE_READLINE_H */
  char buf[INPUT_LINE_BUFFER_LENGHT];

  initialize_readline_history ();

  
#ifdef HAVE_READLINE_READLINE_H
  while ((line = rl_gets ()) != NULL)
    {
  sl_write ("%s", (get_preference(prompt)));
#else
  for (fgets (line, INPUT_LINE_BUFFER_LENGHT, stdin); !feof (stdin);
       fgets (line, INPUT_LINE_BUFFER_LENGHT, stdin))
    {
      if (strlen (line) > INPUT_LINE_BUFFER_LENGHT)
	line[INPUT_LINE_BUFFER_LENGHT - 1] = '\0';
      else
	line[strlen (line) - 1] = '\0';
#endif /* HAVE_READLINE_READLINE_H */
	if (strcmp(line,"") != 0) {
		
      strcpy (buf, line);

      /* parse command line */
      argc = parse_cmdline (buf, argv);

      /* check if is a valid command */
      command_number = check_command (argv[0]);

      if (command_number != SL_SYMBOL_NOT_FOUND)
	Commands[command_number].fnc (argc, argv);
      else
	{
	  /* not a command: try to evaluate... */
	  sl_parser_evaluate (line);
	}
	}
#ifndef HAVE_READLINE_READLINE_H
      sl_write ("%s", (get_preference(prompt)));
#endif /* HAVE_READLINE_READLINE_H */

    }
  return;
}


static int
parse_cmdline (char *buf, char **args_internal)
{
  int argc = 0;

  while (*buf != '\0')
    {
      /* Strip whitespaces */
      while ((*buf == ' ') || (*buf == '\t'))
	*buf++ = (char) NULL;

      /* save the argument */
      *args_internal++ = buf;

      /* skip over args */
      while ((*buf != (char) NULL) && (*buf != ' ') && (*buf != '\t'))
	buf++;

      while ((*buf == ' ') || (*buf == '\t'))
	*buf++ = (char) NULL;
      argc += 1;
    }

  /* ends args_internal with NULL */
  *args_internal = 0;

  return argc;
}
